@@ -1,8 +1,3 @@
-import time
-from mctools.protocol import RCONProtocol, QUERYProtocol, PINGProtocol
-from mctools.packet import RCONPacket, QUERYPacket, PINGPacket
-from mctools.formattertools import FormatterCollection, DefaultFormatter, QUERYFormatter, PINGFormatter
-
 """
 Main Minecraft Connection Clients - Easy to use API for the underlying modules
 Combines the following:
@@ -12,11 +7,21 @@ Combines the following:
 """
 
 
+import time
+
+from mctools.protocol import RCONProtocol, QUERYProtocol, PINGProtocol
+from mctools.packet import RCONPacket, QUERYPacket, PINGPacket
+from mctools.formattertools import FormatterCollection, DefaultFormatter, QUERYFormatter, PINGFormatter
+from mctools.errors import RCONAuthenticationError, RCONMalformedPacketError
+
+__version__ = '1.1.0'
+
+
 class BaseClient(object):
 
     """
     Parent class for Minecraft Client implementations.
-    This class has th3e following formatting constants, which every client inherits:
+    This class has the following formatting constants, which every client inherits:
 
       - BaseClient.RAW - Tells the client to not format any content.
       - BaseClient.REPLACE - Tells the client to replace format characters.
@@ -41,13 +46,29 @@ class BaseClient(object):
 
         return int(time.time())
 
+    def set_timeout(self, timeout):
+
+        """
+        Sets the timeout for the underlying socket object.
+
+        :param timeout: Value in seconds to set the timeout to
+        :type timeout: int
+        """
+
+        # Have the protocol object set the timeout:
+
+        self.proto.set_timeout(timeout)
+
     def start(self):
 
         """
-        Starts the client, and any modules in use.
+        Starts the client, and any protocol objects/formatters in use.
         (Good idea to put a call to Protocol.start() here).
         This raises a 'NotImplementedError' exception,
         as this function should be overridden in the child class.
+
+        :raises:
+            NotImplementedError: If function is called.
         """
 
         raise NotImplementedError("Override this method in child class!")
@@ -55,10 +76,13 @@ class BaseClient(object):
     def stop(self):
 
         """
-        Stops the client, and any modules in use.
+        Stops the client, and any protocol objects/formatters in use.
         (Again, good idea to put a call to Protocol.stop() here).
         This raises a 'NotImplementedError' exception,
         as this function should be overridden in the child class.
+
+        :raises:
+            NotImplementedError: If function is called.
         """
 
         raise NotImplementedError("Override this method in child class!")
@@ -66,11 +90,14 @@ class BaseClient(object):
     def is_connected(self):
 
         """
-        Determine if we are connected to the remote entity.
+        Determine if we are connected to the remote entity,
+        whatever that may be.
         This raises a 'NotImplementedError' exception,
         as this function should be overridden in the child class.
 
         :return: True if connected, False if not.
+        :raises:
+            NotImplementedError: If function is called.
         """
 
         raise NotImplementedError("Override this method in child class!")
@@ -83,6 +110,8 @@ class BaseClient(object):
         as this function should be overridden in the child class.
 
         :param args: Arguments specified, varies from client to client.
+        :raises:
+            NotImplementedError: If function is called.
         """
 
         raise NotImplementedError("Override this method in child class!")
@@ -103,8 +132,6 @@ class BaseClient(object):
 
         """
         All clients MUST have context manager support!
-
-        :return:
         """
 
         raise NotImplementedError("Override this method in child class!")
@@ -130,10 +157,11 @@ class RCONClient(BaseClient):
     RCON client, allows for user to interact with the Minecraft server via the RCON protocol.
 
     :param host: Hostname of the Minecraft server
+    (Can be an IP address or domain name, anything your computer can resolve)
     :type host: str
     :param port: Port of the Minecraft server
     :type port: int
-    :param reqid: Request ID to use, leave as 'None' to generate one based on system time
+    :param reqid: Request ID to use, leave as 'None' to generate an ID based on system time
     :type reqid: int
     :param format_method: Format method to use. You should specify this using the Client constants
     :param format_method: int
@@ -206,17 +234,37 @@ class RCONClient(BaseClient):
 
         return self.auth
 
-    def raw_send(self, reqtype, payload):
+    def raw_send(self, reqtype, payload, frag_check=True):
 
         """
         Creates a RCON packet based off the following parameters and sends it.
+        This function is used for all networking operations.
+
+        We automatically check if a packet is fragmented(We check it's length).
+        If this is the case, then we send a junk packet, and we keep reading packets until the server
+        acknowledges the junk packet. This operation can take some time depending on network speed,
+        so we offer the option to disable this feature, with the risk that their might be stability issues.
+
+        .. warning:
+
+            Use this function at you own risk!
+            You should really call the high level functions,
+            as not doing so could mess up the state/connection of the client.
 
         :param reqtype: Request type
         :type reqtype: int
         :param payload: Payload to send
         :type payload: str
+        :param frag_check: Determines if we should check and handle packet fragmentation.
+
+        .. warning: Disabling fragmentation checks could lead to instability!
+                    Do so at your own risk!
+        :type frag_check: bool
         :return: RCONPacket containing response from server
         :rtype: RCONPacket
+        :raises:
+            RCONAuthenticationError: If the server refuses to server us and we are not authenticated.
+            RCONMalformedPacketError: If the request ID's do not match of the packet is otherwise malformed.
         """
 
         if not self.is_connected():
@@ -233,6 +281,54 @@ class RCONClient(BaseClient):
 
         pack = self.proto.read()
 
+        # Check if our stuff is valid:
+
+        if pack.reqid != self.reqid and self.is_authenticated() and reqtype != self.proto.LOGIN:
+
+            # Client/server ID's do not match!
+
+            raise RCONMalformedPacketError("Client and server request ID's do not match!")
+
+        elif pack.reqid != self.reqid and reqtype != self.proto.LOGIN:
+
+            # Authentication issue!
+
+            raise RCONAuthenticationError("Client and server request ID's do not match! We are not authenticated!")
+
+        # Check if the packet is fragmented(And if we even care about fragmentation):
+
+        if frag_check and pack.length >= self.proto.MAX_SIZE:
+
+            # Send a junk packet:
+
+            self.proto.send(RCONPacket(self.reqid, 0, ''))
+
+            # Read until we get a valid response:
+
+            while True:
+
+                # Get a packet from the server:
+
+                temp_pack = self.proto.read()
+
+                if temp_pack.reqtype == self.proto.RESPONSE and temp_pack.payload == 'Unknown request 0':
+
+                    # Break, we are done here
+
+                    break
+
+                if temp_pack.reqid != self.reqid:
+
+                    # Client/server ID's do not match!
+
+                    raise RCONMalformedPacketError("Client and server request ID's do not match!")
+
+                # Add the packet content to the master pack:
+
+                pack.payload = pack.payload + temp_pack.payload
+
+        # Return our junk:
+
         return pack
 
     def login(self, password):
@@ -276,57 +372,112 @@ class RCONClient(BaseClient):
     def authenticate(self, password):
 
         """
-        Convience function, does the same thing that 'login' does, authenticates you with the RCON server.
+        Convenience function, does the same thing that 'login' does, authenticates you with the RCON server.
 
         :param password: Password to authenticate with
+        :type password: str
         :return: True if successful, False if failure
         :rtype: bool
         """
 
         return self.login(password)
 
-    def command(self, com, no_auth=False):
+    def command(self, com, check_auth=True, format_method=None, return_packet=False, frag_check=True):
 
         """
         Sends a command to the RCON server and gets a response.
 
+        .. note:
+
+            Be sure to authenticate before sending commands to the server!
+            Most servers will simply refuse to talk to you if you do not authenticate.
+
         :param com: Command to send
         :type com: str
-        :param no_auth: Value determining if we should check authentication status.
-        :type no_auth: False
+        :param check_auth: Value determining if we should check authentication status before sending our command
+        :type check_auth: bool
+        :param format_method: Determines the format method we should use. If 'None', then we use the global value
+        You should use the Client constants to define this.
+        :type format_method: int
+        :param return_packet: Determines if we should return the entire packet. If not, return the payload
+        :type return_packet: bool
+        :param frag_check: Determines if we should check and handle packet fragmentation
+
+        .. warning: Disabling fragmentation checks could lead to instability!
+                    Do so at your own risk!
+        :type frag_check: bool
         :return: Response text from server
-        :rtype: str
+        :rtype: str, RCONPacket
+        :raises:
+            RCONAuthenticationError: If we are not authenticated to the RCON server,
+            and authentication checking is enabled. We also raise this if the server refuses to serve us,
+            regardless of weather auth checking is enabled.
+            RCONMalformedPacketError: If the packet we received is broken or is not the correct packet.
         """
 
-        # Getting command packet:
+        # Checking authentication status:
 
-        self.proto.send(RCONPacket(self.reqid, 2, com))
+        if check_auth and not self.is_authenticated():
 
-        # Getting response packet:
+            # Not authenticated, let the user know this:
 
-        pack = self.proto.read()
+            raise RCONAuthenticationError("Not authenticated to the RCON server!")
 
-        if pack.reqid != self.reqid:
+        # Sending command packet:
 
-            # We are not logged in!
+        pack = self.raw_send(self.proto.COMMAND, com, frag_check=frag_check)
 
-            raise Exception("Client is not authenticated!")
+        # Get the formatted content:
+
+        pack = self._format(pack, com, format_method=format_method)
+
+        if return_packet:
+
+            # Return the entire packet:
+
+            return pack
+
+        # Return just the payload
+
+        return pack.payload
+
+    def _format(self, pack, com, format_method=None):
+
+        """
+        Sends incoming data to the formatter.
+
+        :param pack: Packet to format
+        :type pack: RCONPacket
+        :param com: Command issued
+        :type com: str
+        :param format_method: Determines the format method we should use. If 'None', then we use the global value.
+        You should use the Client constants to define this.
+        :type format_method: int
+        :return: Formatted content
+        :rtype: RCONPacket
+        """
+
+        if format_method is None:
+
+            # Use the global format method
+
+            format_method = self.format
 
         # Formatting text:
 
-        if self.format == 'replace' or self.format == 1:
+        if format_method == 'replace' or format_method == 1:
 
             # Replacing format chars
 
             pack.payload = self.formatters.format(pack.payload, com)
 
-        elif self.format == 'clean' or self.format == 2:
+        elif format_method == 'clean' or format_method == 2:
 
             # Removing format chars
 
             pack.payload = self.formatters.clean(pack.payload, com)
 
-        return pack.payload
+        return pack
 
     def __enter__(self):
 
@@ -466,11 +617,16 @@ class QUERYClient(BaseClient):
 
         return pack
 
-    def get_basic_stats(self):
+    def get_basic_stats(self, format_method=None, return_packet=False):
 
         """
         Gets basic stats from the Query server.
 
+        :param format_method: Determines the format method we should use. If 'None', then we use the global value.
+        You should use the Client constants to define this
+        :type format_method: int
+        :param return_packet: Determines if we should return the entire packet. If not, return the payload
+        :type return_packet: bool
         :return: Dictionary of basic stats.
         :rtype: dict
         """
@@ -485,15 +641,28 @@ class QUERYClient(BaseClient):
 
         # Formatting the packet:
 
-        data = self._format(pack)
+        pack.data = self._format(pack, format_method=format_method)
 
-        return data
+        if return_packet:
 
-    def get_full_stats(self):
+            # Return just the packet:
+
+            return pack
+
+        # Return the payload
+
+        return pack.data
+
+    def get_full_stats(self, format_method=None, return_packet=False):
 
         """
         Gets full stats from the Query server.
 
+        :param format_method: Determines the format method we should use. If 'None', then we use the global value.
+        You should use the Client constants to define this.
+        :type format_method: int
+        :param return_packet: Determines if we should return the entire packet. If not, return the payload
+        :type return_packet: bool
         :return: Dictionary of full stats.
         :rtype: dict
         """
@@ -508,31 +677,51 @@ class QUERYClient(BaseClient):
 
         # Formatting the packet:
 
-        data = self._format(pack)
+        pack.data = self._format(pack, format_method=format_method)
 
-        return data
+        if return_packet:
 
-    def _format(self, data):
+            # Return the packet:
+
+            return pack
+
+        # Return the payload
+
+        return pack.data
+
+    def _format(self, data, format_method=None):
 
         """
         Sends the incoming data to the formatter.
 
         :param data: Data to be formatted
         :type data: QUERYPacket
+        :param format_method: Format method to use. If 'None', then we use the global value.
+        :type format_method: int
         :return: Formatted data
         """
 
+        # Check what format type we are using:
+
+        format_type = format_method
+
+        if format_method is None:
+
+            # Use the global format method
+
+            format_type = self.format
+
         # Extracting data from packet:
 
         data = data.data
 
-        if self.format == "replace" or self.format == 1:
+        if format_type == "replace" or format_type == 1:
 
             # Replace format codes with desired values:
 
             data = self.formatters.format(data, "QUERY_PROTOCOL")
 
-        elif self.format in ['clean', 'remove'] or self.format == 2:
+        elif format_type in ['clean', 'remove'] or format_type == 2:
 
             # Remove format codes:
 
@@ -716,11 +905,13 @@ class PINGClient(BaseClient):
 
         return pack, total
 
-    def ping(self):
+    def ping(self, timeout=None):
 
         """
         Pings the Minecraft server and calculates the latency.
 
+        :param timeout: Timeout value used for this operation
+        :type timeout: int
         :return: Time elapsed(in milliseconds).
         :rtype: float
         """
@@ -739,11 +930,16 @@ class PINGClient(BaseClient):
 
         return time_elapsed
 
-    def get_stats(self):
+    def get_stats(self, format_method=None, return_packet=False):
 
         """
         Gets stats from the Minecraft server.
 
+        :param format_method: Determines the format method we should use. If 'None', then we use the global value.
+        You should use the Client constants to define this.
+        :type format_method: int
+        :param return_packet: Determines if we should return the entire packet. If not, return the payload
+        :type return_packet: bool
         :return: Dictionary containing stats.
         :rtype: dict
         """
@@ -766,31 +962,52 @@ class PINGClient(BaseClient):
 
         # Formatting the packet:
 
-        data = self._format(pack)
+        pack.data = self._format(pack, format_method=format_method)
 
-        return data
+        if return_packet:
 
-    def _format(self, pack):
+            # Return the packet:
+
+            return pack
+
+        # Return just the payload
+
+        return pack.data
+
+    def _format(self, pack, format_method=None):
 
         """
         Sends the incoming data to the formatter.
 
         :param pack: Packet to be formatted
         :type pack: PINGPacket
+        :param format_method: Determines the format method we should use. If 'None', then we use the global value.
+        You should use the Client constants to define this.
+        :type format_method: int
         :return: Formatted data.
         """
 
+        # Figure out what format method we are using:
+
+        format_type = format_method
+
+        if format_method is None:
+
+            # Use the global value:
+
+            format_type = self.format
+
         # Extracting the data from the packet:
 
         data = pack.data
 
-        if self.format == "replace" or self.format == 1:
+        if format_type == "replace" or format_type == 1:
 
             # Replace format codes with desired values:
 
             data = self.formatters.format(data, "PING_PROTOCOL")
 
-        elif self.format in ['clean', 'remove'] or self.format == 2:
+        elif format_type in ['clean', 'remove'] or format_type == 2:
 
             # Remove format codes:
 
